/*
  Extension       TalkingClock
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Windows TSE v4.4 upwards,
  Version         v1   5 Jan 2025


  A clock that tells the time   :-)


  This extension tells (!) the time every <interval> minutes
  counting from the whole hour.

  You can configure the voice.
    The voices you can select are the voices that are installed in Windows.
    The voices are presented sorted on language, country, person.

  You can configure the volume.

  You can configure the clock's interval.

  You can configure the clock's own time format.
    For example TSE's time format might be a 24-hour format,
    but for time-telling you would like a 12-hour format.

  Verbal time format
    A special time format "Verbal" was added, which tells the time as it is
    used in spoken language in the selected voice's language and country.
    It starts the time with "It is ..." and the time is told in a longer
    format, especially in Dutch and German which also tell how many minutes
    it is to or past the half hour.
    It does not tell the part of the day, because in practice that would be
    superfluous and become annoying really quickly.
    It supports American English, British English, Dutch, French, German,
    and Spanish.
    Other languages could be added on request.

  A "Test ..." menu option was added to the clock.
    This makes the clock enumerate times according to its settings,
    for each <interval> minutes starting from a <start-time>.


  KNOWN IMPERFECTION

    If there are multiple TSE sessions, then they will all tell the time.
    They do not clash, but wait for each other, as if there is one queue
    for all text-to-speech messages. For now I am OK with this behavior.


  INSTALLATION

    This extension requires that "speak.dll" is installed first.
    It can be found at
      https://eCarlo.nl/tse/Speak.html
    Make sure you have at least version "2024-12-28 v4" of speak:
    This will allow TalkingClock to find all installed Windows voices.

    The installation can just consist of copying speak.dll, I suggest
    to TSE's "mac" directory, which, if you want to copy the speak example
    files too, keeps all of them together.

    Then copy this file "TalkingClock.s" to TSE's "mac" directory, and compile it
    there, for example by opening it there in TSE, and then applying the Macro
    Compile menu.

    Use menu Macro -> Execute "TalkingClock" to configure the clock.

    Optionally start TalkingClock automatically for each TSE session
    by adding "TalkingClock" to the TSE menu Macro -> AutoLoad List.


  TODO
    MUST
    SHOULD
    COULD
      Add an alarm function.
      Add a timer function.
    WONT


  HISTORY

  v0.0.0.1    18 Dec 2024
    Initial version, telling the time each minute.

  v0.0.0.2    19 Dec 2024
    You can now configure the interval at which the clock tells the time.

  v0.0.0.3    21 Dec 2024
    You can now configure a voice by selecting it from a subset of the
    installed Windows voices. No idea what determines the subset.

  v0.0.0.4    24 Dec 2024
    Made the extension's own digital time format configurable.
    Added a "Verbal" time format for American and British English.
    For other languages it does little to nothing, yet.

  v0.0.0.5    27 Dec 2024
    Added a "Test" menu option to the Clock menu to test its settings.

  v0.0.0.6    30 Dec 2024
    Added Dutch, French, German, and Spanish to the "Verbal" speech format.

  v0.0.0.7    31 Dec 2024
    Moved all current settings to the main menu.
    Sorted the list of available voices on language, country, and person.
    Made setting the voice's volume work.

  v0.0.0.8     3 Jan 2025
    The settings were broken. For example, the interval was unchangeable.
    After this update you might need to set some settings again.

  v0.0.0.9     4 Jan 2025
    When selecting a voice, the list of voices is now beautifully formatted.

  v1           5 Jan 2025
    Also say something after configuring the volume and the time format.
    The clock functionality is now complete.
    The alarm and timer functionality are postponed indefinitely.
    Therefore the extension was renamed from AlarmClock to TalkingClock.

*/





/*
  T E C H N I C A L   B A C K G R O U N D   I N F O


  MAKING ONLY ONE TSE SESSION SPEAK THE TIME

  In preliminary research I found these two Windows APIs:
    https://learn.microsoft.com/en-us/windows/win32/api/psapi/nf-psapi-enumprocesses
    https://learn.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-queryfullprocessimagenamea

  Need to be further investigated.


  VOICES ON MY COMPUTER

  The below voice descriptions stay the same independent of the Windows
  language settings.

  Maximum lengths found so far:
    Longest voice description = 49
    Longest person name       =  9
    Longest specific language = 24
    Longest general  language = 10
    Longestn country name     = 14

  MY computer has 9 voices on according to microsoft's old speach api:

    Microsoft Hazel Desktop - English (Great Britain)
    Microsoft Hedda Desktop - German
    Microsoft David Desktop - English (United States)
    Microsoft Zira Desktop - English (United States)
    Microsoft Helena Desktop - Spanish (Spain)
    Microsoft Sabina Desktop - Spanish (Mexico)
    Microsoft Hortense Desktop - French
    Microsoft Elsa Desktop - Italian (Italy)
    Microsoft Maria Desktop - Portuguese(Brazil)

  MY computer has 38 voices according to microsoft's new speach api:

    Microsoft George - English (United Kingdom)
    Microsoft Karsten - German (Switzerland)
    Microsoft Hedda - German (Germany)
    Microsoft Katja - German (Germany)
    Microsoft Stefan - German (Germany)
    Microsoft Catherine - English (Australia)
    Microsoft James - English (Australia)
    Microsoft Linda - English (Canada)
    Microsoft Richard - English (Canada)
    Microsoft Michael - German (Austria)
    Microsoft Hazel - English (United Kingdom)
    Microsoft Susan - English (United Kingdom)
    Microsoft Sean - English (Ireland)
    Microsoft Heera - English (India)
    Microsoft Ravi - English (India)
    Microsoft David - English (United States)
    Microsoft Mark - English (United States)
    Microsoft Zira - English (United States)
    Microsoft Helena - Spanish (Spain)
    Microsoft Laura - Spanish (Spain)
    Microsoft Pablo - Spanish (Spain)
    Microsoft Raul - Spanish (Mexico)
    Microsoft Sabina - Spanish (Mexico)
    Microsoft Caroline - French (Canada)
    Microsoft Claude - French (Canada)
    Microsoft Nathalie - French (Canada)
    Microsoft Guillaume - French (Switzerland)
    Microsoft Hortense - French (France)
    Microsoft Julie - French (France)
    Microsoft Paul - French (France)
    Microsoft Cosimo - Italian (Italy)
    Microsoft Elsa - Italian (Italy)
    Microsoft Jon - Norwegian (Bokml)
    Microsoft Bart - Dutch (Belgium)
    Microsoft Frank - Dutch (Netherlands)
    Microsoft Daniel - Portuguese (Brazil)
    Microsoft Maria - Portuguese (Brazil)
    Microsoft Helia - Portuguese (Portugal)

  Speak first tries the new list, and otherwise tries the old list.
  This should make it compatible with both recent and very old Windows versions.



  PORTUGUESE RESEARCH

  It is eleven o'clock.
  So onze horas.

  It is noon.
   meio-dia.

  It is midnight.
   meia-noite.

  It is one o'clock.
   uma hora.

  It is two o'clock.
  So duas horas.

  It is five pas two.
  So cinco para as duas.

  It is a quarter pas one.
   um e quinze.
   um quarto depois da uma.

  It is a quarter pas two.
   dois e quinze.
   um quarto depois das duas.

*/





// Start of compatibility restrictions and mitigations ...


#ifdef LINUX
  LINUX is not supported. You need a Windows Version of TSE.
#endif

#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE 4.4.
#endif

#ifndef EDITOR_VERSION
   Editor Version is older than TSE 3.0. You need at least TSE 4.4.
#endif

#if EDITOR_VERSION < 4400h
   Editor Version is older than TSE 4.4. You need at least TSE 4.4.
#endif


// End of compatibility restrictions and mitigations.





// Start of Windows APIs ...

dll "<Kernel32.dll>"
  integer proc GetCurrentProcessId(integer void)
end

integer proc get_process_id()
  return(GetCurrentProcessId(0))
end get_process_id

// End of Windows APIs.





// Start of speak implementation ...

// Dec.15.2024 RCH Initial creation
// Dec.24.2024 RCH Added async functions
// Dec.28.2024 RCH Added speak_load_more_voices
// Dec.29.2024 RCH Added speak_get_voice_count, fixes a bug on LoadMoreVoices() were speak_voice_count doesn't change

dll "speak.dll"
    integer proc    speak_initialize            ()

    // For synchronous speech (plays in foreground, blocking)
    integer proc    speak_text                  (string text : cstrval, integer textLength)
    integer proc    speak_text_file             (string filename : cstrval)

    // For asynchronous speech (plays in background, non-blocking)
    integer proc    speak_text_async            (string text : cstrval, integer textLength, integer uid)
    integer proc    speak_text_async_file       (string filename : cstrval, integer uid)
    integer proc    speak_text_async_pause      (integer uid)
    integer proc    speak_text_async_resume     (integer uid)
    integer proc    speak_text_async_stop       (integer uid)

    // Setters and getters
    integer proc    speak_set_volume            (integer volume_0_to_100)
    integer proc    speak_set_voice             (integer voiceNum)
    integer proc    speak_get_voice             ()
    integer proc    speak_get_voice_count       ()                                                      // Dec.29.2024 RCH added
    integer proc    speak_get_volume            ()
    integer proc    speak_get_voice_info        (integer voiceNum, var string output255 : strptr)
    integer proc    speak_get_api_info          (integer apiNum, var string output255 : strptr)         // Dec.24.2024 RCH added
    integer proc    speak_write_out_api_info    (string outputFilename : cstrval)                       // Dec.24.2024 RCH added
    integer proc    speak_load_more_voices      (string registryPath : cstrval)                         // Dec.28.2024 RCH added
end

integer   speak_voice_count           = 0
integer   speak_last_voice            = 0
integer   is_speak_initialized        = FALSE
integer   speak_show_no_msgboxes      = FALSE
string    speak_not_initialized[34]   = "SPEAK.DLL count not be initialized"

proc Initialize()
    if is_speak_initialized == FALSE
        speak_voice_count = speak_initialize()
        if speak_voice_count >= 1
            // We have at least one voice to work with
            speak_set_voice(0)
            speak_set_volume(50)
            is_speak_initialized = TRUE
        endif
    endif
end

proc SpeakNotInitialized()

    string text[47] = "SPEAK.DLL:  The engine could not be initialized"

    if speak_show_no_msgboxes == FALSE
        MsgBox(text)
    else
        Message(text)
    endif
end

proc Speak(string text)
    Initialize()
    if is_speak_initialized
        speak_text(text, Length(text))
        return()
    endif
    SpeakNotInitialized()
end

proc SpeakAsync(string text, integer uid)
    Initialize()
    if is_speak_initialized
        speak_text_async(text, Length(text), uid)
        return()
    endif
    SpeakNotInitialized()
end

proc SpeakAsyncPause(integer uid)
    Initialize()
    if is_speak_initialized
        speak_text_async_pause(uid)
        return()
    endif
    SpeakNotInitialized()
end

proc SpeakAsyncResume(integer uid)
    Initialize()
    if is_speak_initialized
        speak_text_async_resume(uid)
        return()
    endif
    SpeakNotInitialized()
end

proc SpeakAsyncStop(integer uid)
    Initialize()
    if is_speak_initialized
        speak_text_async_stop(uid)
        return()
    endif
    SpeakNotInitialized()
end

integer proc SetVolume(integer volume)
    Initialize()
    if is_speak_initialized
        volume = speak_set_volume(volume)
        return(volume)
    endif
    SpeakNotInitialized()
    return(-1)
end

integer proc GetVolume()
    Initialize()
    if is_speak_initialized
        return(speak_get_volume())
    endif
    SpeakNotInitialized()
    return(-1)
end

proc SetVoice(integer voice)
    Initialize()
    if is_speak_initialized
        if voice < speak_voice_count
            speak_last_voice = voice
            speak_set_voice(voice)
            return()
        endif
        Message("SPEAK.DLL:  Voice " + Str(voice) + " is invalid.  Max voice is " + Str(speak_voice_count - 1))
        return()
    endif
    SpeakNotInitialized()
end

// Call with -1 to use the current voice
string proc GetVoice(integer voice)

    integer len
    string output255[255] = Format(" " : 255)

    Initialize()
    if is_speak_initialized
        if voice < 0
            voice = speak_get_voice()
        endif
        if voice < speak_voice_count
            len = speak_get_voice_info(voice, output255)
            if len > 0
                return(LeftStr(output255, len))
            endif
            return("An error occurred while retrieving voice #" + Str(voice) + ".  Error #" + Str(Abs(len)))
        else
            return("Invalid voice #" + Str(voice) + ", must be 0 thru " + Str(speak_voice_count - 1))
        endif
    endif
    SpeakNotInitialized()
    return(speak_not_initialized)

end

integer proc LoadMoreVoices(string registryPath)

    integer count, before

    Initialize()
    if is_speak_initialized
        count = speak_load_more_voices(registryPath)
        before = speak_voice_count
        speak_voice_count = speak_get_voice_count()
        Message("SPEAK.DLL:  " + Str(count) + " voice" + iif(count <> 1, "s", "") + " loaded.  " + iif(count <> 0, "Was " + Str(before) + ", now ", "") + Str(speak_voice_count) + " voice" + iif(count <> 1, "s", "") + " available.")
    else
        count = -1
        SpeakNotInitialized()
    endif
    return(count)

end

proc SpeakChooseVoice()

    integer ok
    string text[3] = ""

    Initialize()
    if is_speak_initialized
        ok = AskNumeric("Please choose a voice number 0 to " + Str(speak_voice_count - 1) + ":", text)
        if ok and Val(text) >= 0 and Val(text) < speak_voice_count
            SetVoice(Val(text))
            Message("SPEAK.DLL:  Voice is now set to #" + text + " of " + Str(speak_voice_count - 1) + " " + Chr(34) + GetVoice(Val(text)) + Chr(34))
            return()
        endif
        Message("SPEAK.DLL:  Voice #" + text + " is invalid.  Must be 0 to " + Str(speak_voice_count -1))
        return()
    endif
    SpeakNotInitialized()

end

proc SpeakPrevVoice()

    integer voice
    string text[255] = ""

    Initialize()
    if is_speak_initialized
        voice = speak_get_voice()
        if voice > 0
            SetVoice(voice - 1)
        else
            SetVoice(speak_voice_count - 1)
        endif
        voice = speak_get_voice()
        text  = "Current voice is set to #" + Str(voice) + " of " + Str(speak_voice_count - 1) + ".  " + GetVoice(voice) + "."
        Message("SPEAK.DLL:  " + text)
        SpeakAsync(text, 10)
        return()
    endif
    SpeakNotInitialized()

end

proc SpeakNextVoice()

    integer voice
    string text[255] = ""

    Initialize()
    if is_speak_initialized
        voice = speak_get_voice()
        if voice < speak_voice_count - 1
            SetVoice(voice + 1)
        else
            SetVoice(0)
        endif
        voice = speak_get_voice()
        text  = "Current voice is set to #" + Str(voice) + " of " + Str(speak_voice_count - 1) + ".  " + GetVoice(voice) + "."
        Message("SPEAK.DLL:  " + text)
        SpeakAsync(text, 10)
        return()
    endif
    SpeakNotInitialized()

end

proc SpeakChooseText()

    integer ok
    string text[255] = "This is " + GetVoice(speak_last_voice)

    Initialize()
    if is_speak_initialized
        ok = Ask("Please choose some text to speak:", text)
        if ok
            Message("SPEAK.DLL:  " + text)
            SpeakAsync(text, 10)
        endif
    endif
end

proc SpeakChooseVolume()

    integer ok, oldVolume
    string text[3] = ""

    Initialize()
    if is_speak_initialized
        text = Str(speak_get_volume())
        ok = AskNumeric("Please choose a volume 0 to 100", text)
        if ok
            if Val(text) >= 0 and Val(text) <= 100
                oldVolume = SetVolume(Val(text))
                Message("SPEAK.DLL:  Volume set to " + text + ", was " + Str(oldVolume))
            else
                Message("SPEAK.DLL:  Volume not set.  Must be in range 0..100.  Is currently " + Str(oldVolume))
            endif
            return()
        endif
        Message("SPEAK.DLL:  The value " + text + " is invalid.  Must be 0 to 100")
        return()
    endif
    SpeakNotInitialized()

end

string proc GetApi(integer api)

    integer len
    string output255[255] = Format(" " : 255)

    Initialize()
    if is_speak_initialized
        len = speak_get_api_info(api, output255)
        if len > 0
            return(LeftStr(output255, len))
        endif
        return("An error occurred while retrieving API #" + Str(api) + ".  Error #" + Str(len))
    endif
    SpeakNotInitialized()
    return(speak_not_initialized)

end

string proc WriteApiToFile(string outputFilename)
    Initialize()
    if is_speak_initialized
        if speak_write_out_api_info(outputFilename) >= 0
            Message("SPEAK.DLL:  API information written to " + outputFilename)
            return("Written")
        endif
        Message("SPEAK.DLL:  An error occurred attemtping to write API information to " + outputFilename)
        return("An error occurred while attempting to write to file: " + outputFilename)
    endif
    SpeakNotInitialized()
    return(speak_not_initialized)
end

string proc InsertApiIntoEditor()

    integer api, len
    string output255[255] = Format(" " : 255)

    Initialize()
    if is_speak_initialized
        api = 0
        loop
            len = speak_get_api_info(api, output255)
            if len < 0
                break
            endif
            if len <> 0
                AddLine(LeftStr(output255, len))
            endif
            api = api + 1
        endloop
        AddLine("Found " + Str(api) + " API function" + iif(api <> 1, "s", ""))
        AddLine()
        return(GetText(1, CurrLineLen()))
    endif
    SpeakNotInitialized()
    return(speak_not_initialized)

end

string proc InsertVoicesIntoEditor()

    integer voice, len
    string output255[255] = Format(" " : 255)

    Initialize()
    if is_speak_initialized
        voice = 0
        loop
            len = speak_get_voice_info(voice, output255)
            if len < 0
                break
            endif
            if len <> 0
                AddLine("Voice #" + Str(voice) + " -- " + LeftStr(output255, len))
            endif
            voice = voice + 1
        endloop
        AddLine("Found " + Str(voice) + " voice" + iif(voice <> 1, "s", ""))
        AddLine()
        return(GetText(1, CurrLineLen()))
    endif
    SpeakNotInitialized()
    return(speak_not_initialized)

end

// End of speak implementation.



// Start of speak compensation ...

proc dummy()
  if FALSE
    dummy()
    GetApi(0)
    GetVoice(0)
    GetVolume()
    Initialize()
    InsertApiIntoEditor()
    InsertVoicesIntoEditor()
    LoadMoreVoices('')
    SetVoice(0)
    SetVolume(0)
    Speak('')
    SpeakAsync('', 0)
    SpeakAsyncPause(0)
    SpeakAsyncResume(0)
    SpeakAsyncStop(0)
    SpeakChooseText()
    SpeakChooseVoice()
    SpeakChooseVolume()
    SpeakNextVoice()
    SpeakPrevVoice()
    WriteApiToFile('')
  endif
end dummy
// End of speak compensation.





// Global constants and semi-constants
#define CLOCK_DELAY                   21  // Tested: 21 calls ~> 5 seconds.
integer CLOCK_ID                    = 0
string  CLOCK_INTERVAL_TO_HIST [33] = 'Off 1 2 3 4 5 6 10 12 15 20 30 60'
#define COLON                         Asc(':')
integer MACRO_ID                    = 0
string  MACRO_NAME   [MAXSTRINGLEN] = ''
#define MAX_VOICE_LEN                 49


// Global variables
integer cfg_clock_format                = 5
integer cfg_clock_interval              = 1
string  cfg_voice_fqn   [MAX_VOICE_LEN] = ''
integer cfg_voice_id                    = 0
integer cfg_volume                      = 100
integer clock_format_history            = 0
integer clock_interval_hist             = 1
integer clock_timer                     = CLOCK_DELAY
string  prev_clock_time             [5] = ''
integer stop_main_menu                  = FALSE
integer testing                         = FALSE
integer test_hours                      = 0
integer test_minutes                    = 0
integer voice_ok                        = TRUE


integer proc write_profile_int(string  section_name,
                               string  item_name,
                               integer item_value)
  integer ok = WriteProfileInt(section_name,
                               item_name,
                               item_value)
  if not ok
    Alarm()
    MsgBox(MACRO_NAME, 'Could not write item "' + item_name +
                       '" to configuration file "tse.ini".')
  endif

  return(ok)
end write_profile_int


integer proc write_profile_str(string section_name,
                               string item_name,
                               string item_value)
  integer ok = WriteProfileStr(section_name,
                               item_name,
                               item_value)
  if not ok
    Alarm()
    MsgBox(MACRO_NAME, 'Could not write item "' + item_name +
                       '" to configuration file "tse.ini".')
  endif

  return(ok)
end write_profile_str


string proc get_country()
  integer p                      = Pos('-', cfg_voice_fqn)
  integer p_fr                   = 0
  integer p_to                   = 0
  string  result [MAX_VOICE_LEN] = ''
  if p
    result = Trim(cfg_voice_fqn[(p + 1) .. MAXSTRINGLEN])
    p_fr   = Pos('(', result)
    p_to   = Pos(')', result)
    if p_fr
      if p_to - p_fr > 1
        result = result[(p_fr + 1) .. (p_to - 1)]
      else
        result = result[(p_fr + 1) .. MAXSTRINGLEN]
      endif
    else
      result = ''
    endif
  endif
  return(result)
end get_country


string proc get_language()
  integer p                      = Pos('-', cfg_voice_fqn)
  string  result [MAX_VOICE_LEN] = cfg_voice_fqn
  if p
    result = Trim(result[(p + 1) .. MAXSTRINGLEN])
  endif
  p = Pos('(', result)
  if p
    result = Trim(result[1 .. (p - 1)])
  endif
  return(result)
end get_language


string proc verbalize_time(string digital_time)
  string  country    [MAX_VOICE_LEN] = ''
  integer hours                      = Val(digital_time[1: 2])
  string  language   [MAX_VOICE_LEN] = ''
  integer minutes                    = Val(digital_time[4: 2])
  integer org_hours                  = 0
  string  verbal_time [MAXSTRINGLEN] = ''

  org_hours = hours
  language  = get_language()

  case language

    when 'Dutch'
      verbal_time = 'Het is'

      if minutes >= 16
        hours = hours + 1
      endif
      hours = hours mod 12
      if hours == 0
        hours = 12
      endif

      case minutes
        when 0
          if org_hours == 0
            verbal_time = Format(verbal_time; 'middernacht.')
          else
            verbal_time = Format(verbal_time; hours; 'uur.')
          endif
        when 30
          verbal_time = Format(verbal_time; 'half'; hours, '.')
        when 15
          verbal_time = Format(verbal_time; 'kwart over'; hours, '.')
        when 45
          verbal_time = Format(verbal_time; 'kwart voor'; hours, '.')
        when 1 .. 14
          verbal_time = Format(verbal_time; minutes; 'over'; hours, '.')
        when 16 .. 29
          verbal_time = Format(verbal_time; 30 - minutes; 'voor half'; hours, '.')
        when 31 .. 44
          verbal_time = Format(verbal_time; minutes - 30; 'over half'; hours, '.')
        when 46 .. 59
          verbal_time = Format(verbal_time; 60 - minutes; 'voor'; hours, '.')
      endcase

    when 'English'
      country = get_country()

      if (country in 'Canada', 'United States')
        if minutes == 0
          if     hours == 12
            verbal_time = 'It is noon.'
          elseif hours == 0
            verbal_time = 'It is midnight.'
          else
            verbal_time = Format('It is'; hours mod 12; "o'clock.")
          endif
        else
          if hours >= 13
            hours = hours mod 12
          endif
          verbal_time = Format('It is'; hours, ':', minutes: 2: '0', '.')
        endif
      else
        // Other "English" countries favor the British way of telling time.
        verbal_time = 'It is'
        if minutes == 0
          if     hours == 12
            verbal_time = verbal_time + ' noon.'
          elseif hours == 0
            verbal_time = verbal_time + ' midnight.'
          else
            verbal_time = verbal_time + ' ' +
                          Str(iif(hours < 13, hours, hours mod 12)) +
                          " o'clock."
          endif
        else
          if minutes == 30
            verbal_time = verbal_time + ' half'
          elseif (minutes in 15, 45)
            verbal_time = verbal_time + ' a quarter'
          else
            if (minutes in 1 .. 29)
              verbal_time = verbal_time + ' ' + Str(minutes)
            elseif (minutes in 31 .. 59)
              verbal_time = verbal_time + ' ' + Str(60 - minutes)
            endif
          endif
          if minutes <= 30
            verbal_time = verbal_time + ' past'
          else
            verbal_time = verbal_time + ' to'
          endif
          if minutes > 30
            hours = hours + 1
          endif
          if (hours in 0, 24)
            verbal_time = verbal_time + ' midnight.'
          else
            verbal_time = verbal_time + ' ' +
                          iif(hours <= 12, Str(hours), Str(hours mod 12)) + '.'
          endif
        endif
      endif

    when 'French'
      if minutes >= 31
        hours = hours + 1
      endif
      hours = hours mod 12
      if hours == 0
        hours = 12
      endif

      verbal_time = 'Il est'

      case minutes
        when 0
          if org_hours == 0
            verbal_time = Format(verbal_time; 'minuit.')
          else
            verbal_time = Format(verbal_time; hours; 'heurs.')
          endif
        when 30
          verbal_time = Format(verbal_time; hours; 'heurs et demie.')
        when 15
          verbal_time = Format(verbal_time; hours; 'et quart.')
        when 45
          verbal_time = Format(verbal_time; hours; 'moins le quart.')
        when 1 .. 14, 16 .. 29
          verbal_time = Format(verbal_time; hours; minutes, '.')
        when 31 .. 44, 46 .. 59
          verbal_time = Format(verbal_time; hours; 'moins'; 60 - minutes, '.')
      endcase

    when 'German'
      verbal_time = 'Es ist'

      if minutes >= 16
        hours = hours + 1
      endif
      hours = hours mod 12
      if hours == 0
        hours = 12
      endif

      case minutes
        when 0
          verbal_time = Format(verbal_time; hours; 'Uhr.')
        when 30
          verbal_time = Format(verbal_time; 'halb'; hours; '.')
        when 15
          verbal_time = Format(verbal_time; 'Viertel nach'; hours; '.')
        when 45
          verbal_time = Format(verbal_time; 'Viertel vor'; hours; '.')
        when 1 .. 14
          verbal_time = Format(verbal_time; minutes; 'nach'; hours; '.')
        when 16 .. 29
          verbal_time = Format(verbal_time; 30 - minutes; 'vor halb'; hours ; '.')
        when 31 .. 44
          verbal_time = Format(verbal_time; minutes - 30; 'nach halb'; hours; '.')
        when 46 .. 59
          verbal_time = Format(verbal_time; 60 - minutes; 'vor'; hours; '.')
      endcase

    when 'Spanish'
      if minutes >= 31
        hours = hours + 1
      endif
      hours = hours mod 12
      if hours == 0
        hours = 12
      endif

      verbal_time = iif(hours == 1, 'Es la', 'Son las')

      case minutes
        when 0
          verbal_time = Format(verbal_time; hours; 'en punto.')
        when 30
          verbal_time = Format(verbal_time; hours; 'y media.')
        when 15
          verbal_time = Format(verbal_time; hours; 'y cuarto.')
        when 45
          verbal_time = Format(verbal_time; hours; 'menos cuarto.')
        when 1 .. 14, 16 .. 29
          verbal_time = Format(verbal_time; hours; 'y'; minutes, '.')
        when 31 .. 44, 46 .. 59
          verbal_time = Format(verbal_time; hours; 'menos'; 60 - minutes, '.')
      endcase

    otherwise
      if     hours == 0
        hours = 12
      elseif hours >= 13
        hours = hours mod 12
      endif
      verbal_time = Format(hours, ':', minutes: 2: '0')
  endcase

  return(verbal_time)
end verbalize_time


string proc get_test_time()
  integer time_format    = 0
  integer time_separator = 0
  string  time       [8] = ''

  case cfg_clock_format
    when 2, 3
      time_format    = 2
      time_separator = COLON
    when 4, 5
      time_format    = 1
      time_separator = COLON
    otherwise
      time_format    = Query(TimeFormat)
      time_separator = Query(TimeSeparator)
  endcase

  case time_format
    when 2
      time = Format((test_hours mod 12):2:' ',
                    Chr(time_separator),
                    test_minutes:2:'0',
                    ' ',
                    iif(test_hours < 12, 'am', 'pm'))
    when 3
      time = Format(test_hours:2:'0',
                    Chr(time_separator),
                    test_minutes:2:'0',
                    Chr(time_separator),
                    '00')
    when 4
      time = Format((test_hours mod 12):2:'0',
                    Chr(time_separator),
                    test_minutes:2:'0',
                    ' ',
                    iif(test_hours < 12, 'am', 'pm'))
    otherwise
      time = Format(test_hours:2:' ',
                    Chr(time_separator),
                    test_minutes:2:'0',
                    Chr(time_separator),
                    '00')
  endcase

  return(time)
end get_test_time



string proc get_real_time()
  integer old_TimeFormat       = 0
  integer old_TimeSeparatorSet = 0
  string  time             [8] = ''

  case cfg_clock_format
    when 2, 3
      old_TimeFormat       = Set(TimeFormat   , 2    )
      old_TimeSeparatorSet = Set(TimeSeparator, COLON)
    when 4, 5
      old_TimeFormat       = Set(TimeFormat   , 1    )
      old_TimeSeparatorSet = Set(TimeSeparator, COLON)
  endcase

  time = GetTimeStr()

  case cfg_clock_format
    when 2, 3, 4, 5
      Set(TimeFormat   , old_TimeFormat      )
      Set(TimeSeparator, old_TimeSeparatorSet)
  endcase

  return(time)
end get_real_time


string proc get_time()
  string time [8] = ''

  if testing
    time = get_test_time()
  else
    time = get_real_time()
  endif

  return(time)
end get_time


proc tell_time()
  string  time [MAXSTRINGLEN] = ''

  time = get_time()

  if time[6: 1] == ':'
    time = time[1: 5]
  endif

  if cfg_clock_format == 5
    time = verbalize_time(time)
  endif

  if testing
    if cfg_clock_format == 5
      Message(time)
    else
      Message('Testing'; time)
    endif

    if TRUE
      Speak(time)
    else
      // Debug instead
      if not Pos('unnamed', CurrFilename())
        NewFile()
      endif
      AddLine(time)
      UpdateDisplay(_ALL_WINDOWS_REFRESH_)
      Delay(1)
    endif
  else
    SpeakAsync(time, CLOCK_ID)
  endif
end tell_time


proc test_time_telling()
  integer increment            = 0
  integer increment_history    = GetFreeHistory(MACRO_NAME + ':TestIncrement')
  string  increment_answer [3] = ''
  integer ok                   = TRUE
  integer old_cursor           = Query(Cursor)
  string  start_time       [5] = ''
  integer start_time_history   = GetFreeHistory(MACRO_NAME + ':TestStartTime')

  if GetHistoryStr(start_time_history, 1) == ''
    start_time = '2230'
  endif

  ok = Ask('Tell the time in 5-minute increments from (h or hh or hmm or hhmm):',
           start_time, start_time_history)

  if ok
    start_time = Str(Val(start_time))

    if Length(start_time) <= 2
      test_hours   = Abs(Val(start_time)) mod 24
      test_minutes = 0
    else
      test_hours   = Abs(Val(start_time[1: Length(start_time) - 2])) mod 24
      test_minutes = Abs(Val(start_time[Length(start_time) - 1: 2])) mod 60
    endif

    if GetHistoryStr(increment_history, 1) == ''
      increment_answer = '5'
    endif

    ok = Ask('Increment (minutes):', increment_answer, increment_history)
  endif

  if ok
    increment = Abs(Val(increment_answer))
    increment = iif(increment <= 60, increment, increment mod 60)

    Set(Cursor, OFF)
    Message('Press <Escape> to stop ...')
    Delay(36)
    testing = TRUE

    repeat
      tell_time()

      test_minutes = test_minutes + increment

      if test_minutes >= 60
        test_minutes =  test_minutes    mod 60
        test_hours   = (test_hours + 1) mod 24
      endif
    until KeyPressed()
      and GetKey()
      and Query(Key) == <Escape>

    Set(Cursor, old_cursor)
  endif

  testing              = FALSE
  stop_main_menu = FALSE
  UpdateDisplay(_STATUSLINE_REFRESH_)
end test_time_telling


// Time is something this tool can tell any configured language.
// We pick the longest format of spoken time that is available.
proc say_something_after_cfg_change()
  integer old_clock_format = cfg_clock_format
  tell_time()
  cfg_clock_format = old_clock_format
end say_something_after_cfg_change


proc idle()
  string curr_clock_time [5] = ''

  if cfg_clock_interval
    if clock_timer
      clock_timer = clock_timer - 1
    else
      clock_timer     = CLOCK_DELAY
      curr_clock_time = GetTimeStr()

      if  curr_clock_time <> prev_clock_time
      and Val(curr_clock_time[4: 2]) mod cfg_clock_interval == 0
        tell_time()
        prev_clock_time = curr_clock_time
      endif
    endif
  endif
end idle


menu clock_interval_menu()
  title   = 'Clock -> Interval (minutes)'
  history = clock_interval_hist

  '&Off',,, 'Do not tell the time'
    '&1',,, 'Tell the time every minute'
    '&2',,, 'Tell the time every 2 minutes'
     '3',,, 'Tell the time every 3 minutes'
     '4',,, 'Tell the time every 4 minutes'
    '&5',,, 'Tell the time every 5 minutes'
     '6',,, 'Tell the time every 6 minutes'
   '1&0',,, 'Tell the time every 10 minutes'
    '12',,, 'Tell the time every 12 minutes'
    '15',,, 'Tell the time every quarter hour'
    '20',,, 'Tell the time every 20 minutes'
   '&30',,, 'Tell the time every half hour'
   '&60',,, 'Tell the time every hour'
end clock_interval_menu


proc configure_clock_interval()
  integer i = 0

  for i = 1 to NumTokens(CLOCK_INTERVAL_TO_HIST, ' ')
    if Val(GetToken(CLOCK_INTERVAL_TO_HIST, ' ', i)) == cfg_clock_interval
      clock_interval_hist = i
      break
    endif
  endfor

  clock_interval_menu()

  if MenuOption()
    cfg_clock_interval = Val(MenuStr(clock_interval_menu, MenuOption()))
    write_profile_int(MACRO_NAME + ':ClockSettings', 'Interval', cfg_clock_interval)
  endif

  stop_main_menu = FALSE
end configure_clock_interval


proc configure_voice()
  string  country_name  [MAXSTRINGLEN] = ''
  integer country_pos                  = 0
  integer downgrade_pos                = 0
  integer i                            = 0
  string  language_name [MAXSTRINGLEN] = ''
  integer language_pos                 = 0
  integer old_list_width               = 0
  integer new_list_width               = 0
  integer old_id                       = GetBufferId()
  string  person_name   [MAXSTRINGLEN] = ''
  integer person_pos                   = 0

  GotoBufferId(MACRO_ID)
  EmptyBuffer()

  while LeftStr(GetVoice(i), 15) <> 'Invalid voice #'
    AddLine(GetVoice(i))
    i = i + 1
  endwhile

  if NumLines()
    //  Microsoft's full voice name format is "Name - Language (Country)".
    //  In Microsoft's old API the "(Country)" part is sometimes omitted.
    //  The below new temporary list format should downgrade gracefully if the
    //  full voice name format does not adhere to the above conventions.
    //
    //  To the right of the list create aligned columns for language, country,
    //  and person, and sort the combined list on the left list.
    old_list_width = LongestLineInBuffer()

    downgrade_pos  = LongestLineInBuffer() + 1
    for i = 1 to NumLines()
      GotoLine(i)
      GotoPos(downgrade_pos)
      InsertText('0')
    endfor

    language_pos = LongestLineInBuffer() + 1
    for i = 1 to NumLines()
      GotoLine(i)
      language_name = Trim(GetToken(GetText(1, old_list_width), '-', 2))
      if language_name == ''
        GotoPos(downgrade_pos)
        InsertText('1', _OVERWRITE_)
      else
        language_name = Trim(GetToken(language_name, '(', 1))
        GotoPos(language_pos)
        InsertText(language_name)
      endif
    endfor

    country_pos = LongestLineInBuffer() + 2
    for i = 1 to NumLines()
      GotoLine(i)
      if not Val(GetText(downgrade_pos, 1))
        country_name = Trim(GetToken(GetText(1, old_list_width), '-', 2))
        if Pos('(', country_name)
          country_name = Trim(GetToken(country_name, '(', 2))
          if Pos(')', country_name) == Length(country_name)
            country_name = Trim(GetToken(country_name, ')', 1))
          endif
        else
          country_name = ''
        endif
        GotoPos(country_pos)
        InsertText(country_name)
      endif
    endfor

    person_pos = LongestLineInBuffer() + 2
    for i = 1 to NumLines()
      GotoLine(i)
      if Val(GetText(downgrade_pos, 1))
        //  If downgraded, put full voice name in column for person name.
        InsertText(Trim(GetText(1, old_list_width)))
      else
        person_name = Trim(GetToken(GetText(1, old_list_width), '-', 1))
        //  In the list remove "Microsoft" from the person name.
        person_name = StrReplace('Microsoft', person_name, '', 'i')
        //  In case of the old voice API also remove "Desktop" from the person
        //  name.
        person_name = StrReplace('Desktop', person_name, '', 'i')
        GotoPos(person_pos)
        InsertText(person_name)
      endif
    endfor

    PushBlock()

    //  Switch the side-by-side lists,
    //  deleting the downgrade-column,
    //  and adding a space after the new list.
    new_list_width = LongestLineInBuffer() - old_list_width - 1
    MarkColumn(1, language_pos , NumLines(), LongestLineInBuffer())
    Cut()
    MarkColumn(1, downgrade_pos, NumLines(), downgrade_pos)
    KillBlock()
    lReplace('^', ' ', 'gnx')
    BegFile()
    Paste()
    lReplace(' #$', '', 'gnx')

    //  Sort on the new list, which is now on the left side.
    MarkColumn(1, 1, NumLines(), new_list_width)
    Sort(_IGNORE_CASE_)

    //  Move the cursor to the old selected voice's line.
    MarkColumn(1, new_list_width + 2, NumLines(), LongestLineInBuffer())
    BegFile()
    lFind(cfg_voice_fqn, '^gl$')
    BegLine()

    PopBlock()

    //  Present the left list to the user.
    if List('Select a voice',
            new_list_width + 1)
      cfg_voice_fqn = Trim(GetText(new_list_width + 2, MAXSTRINGLEN))

      // Find the voice id of the selected voice description.
      for i = 0 to NumLines() - 1
        if GetVoice(i) == cfg_voice_fqn
          cfg_voice_id = i
          break
        endif
      endfor

      write_profile_str(MACRO_NAME + ':GeneralSettings',
                        'Voice',
                        cfg_voice_fqn)
      SetVoice(cfg_voice_id)
    endif
    say_something_after_cfg_change()
  else
    MsgBox(MACRO_NAME,
           'There are no text-to-speech voices available in your Windows settings.',
           _OK_)
  endif

  GotoBufferId(old_id)
  stop_main_menu = FALSE
end configure_voice


proc configure_clock_volume()
  string answer [4] = Str(cfg_volume)

  if AskNumeric('Volume [0 - 100]:', answer)
    if  Trim(answer) <> ''
    and (Val(answer) in 0 .. 100)
      cfg_volume = Val(answer)
      SetVolume(cfg_volume)
      write_profile_int(MACRO_NAME + ':GeneralSettings', 'Volume', cfg_volume)
    else
      Warn('"', Trim(answer), '" is an invalid volume setting.')
    endif
  endif

  say_something_after_cfg_change()
  stop_main_menu = FALSE
end configure_clock_volume


string proc get_clock_format()
  string result [18] = ''

  case cfg_clock_format
    when 2
      result = '1:00 - 12:00'
    when 3
      result = '1:00 am - 12:00 pm'
    when 4
      result = '1:00 - 23:00'
    when 5
      result = 'Verbal'
    otherwise
      result = "TSE's time format"
  endcase

  return(result)
end get_clock_format


menu clock_format_menu()
  title   = 'Clock -> Format'
  history = clock_format_history

  "&TSE's time format"
  '1:00    - &12:00'
  '1:00 am - 1&2:00 pm'
  '0:00    - 2&3:59'
  '&Verbal'
end clock_format_menu


proc configure_clock_format()
  clock_format_history = cfg_clock_format

  clock_format_menu()

  if MenuOption()
    cfg_clock_format = MenuOption()
    write_profile_int(MACRO_NAME + ':ClockSettings', 'Format', cfg_clock_format)
  endif

  say_something_after_cfg_change()
  stop_main_menu = FALSE
end configure_clock_format


menu main_menu()
  title = 'TalkingClock'
  history

  'General Settings',,_MF_DIVIDE_

  '&Voice'
    [Trim(cfg_voice_fqn):49],
    configure_voice(),,
    "Select an installed Windows voice (Sorted on language, country, person)"

  'Vol&ume                               (percentage)'
    [Abs(cfg_volume):3],
    configure_clock_volume(),,
    "Set the clock's volume to a percentage from 0 to 100"

  'Clock Settings',,_MF_DIVIDE_

  '&Interval                                (minutes)'
    [iif(cfg_clock_interval, Str(cfg_clock_interval), 'Off'):3],
    configure_clock_interval(),,
    'Every how many minutes should the clock tell the time'

  '&Format'
    [get_clock_format():18],
    configure_clock_format(),,
    'Set clock format to "0:00-23:59", "0:00-11:59", "0:00am-11:59pm", or "Verbal".'

  '&Test ...',
    test_time_telling(),,
    "Continuously tell a test time from a start-time in <interval>-minute increments"
end main_menu


proc configure_main()
  repeat
    stop_main_menu = TRUE
    main_menu()
  until stop_main_menu

  stop_main_menu = stop_main_menu

end configure_main


proc get_voice()
  integer i                    = 0
  string  voice [MAXSTRINGLEN] = ''

  voice = GetProfileStr(MACRO_NAME + ':GeneralSettings', 'Voice', '')

  if voice <> ''
    while LeftStr(GetVoice(i), 15) <> 'Invalid voice #'
      if voice == GetVoice(i)
        cfg_voice_id  = i
        cfg_voice_fqn = voice
        SetVoice(cfg_voice_id)
        break
      endif
      i = i + 1
    endwhile

    if cfg_voice_fqn == ''
      MsgBox(MACRO_NAME,
             Format('The configured voice',
                    Chr(13),
                    '  ', cfg_voice_fqn,
                    Chr(13),
                    'is not a Windows text-to-speech voice.'),
             _OK_)
    endif
  endif

  if cfg_voice_fqn == ''
    voice = GetVoice(0)

    if LeftStr(voice, 15) == 'Invalid voice #'
      voice_ok = FALSE
      MsgBox(MACRO_NAME,
             'Abort: No Windows text-to-speech voice available.',
             _OK_)
      PurgeMacro(CurrMacroFilename())
    else
      cfg_voice_id  = 0
      cfg_voice_fqn = voice
      SetVoice(cfg_voice_id)
    endif
  endif

end get_voice


proc WhenPurged()
  AbandonFile(MACRO_ID)
end WhenPurged


proc WhenLoaded()
  integer org_id = GetBufferId()

  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)

  get_voice()

  if voice_ok
    cfg_volume         = GetProfileInt(MACRO_NAME + ':GeneralSettings', 'Volume'  , 100)
    cfg_clock_interval = GetProfileInt(MACRO_NAME + ':ClockSettings'  , 'Interval',   1)
    cfg_clock_format   = GetProfileInt(MACRO_NAME + ':ClockSettings'  , 'Format'  ,   5)

    cfg_volume         = iif((cfg_volume         in 0 .. 100), cfg_volume        , 100)
    cfg_clock_interval = iif((cfg_clock_interval in 1 ..  60), cfg_clock_interval,   1)
    cfg_clock_format   = iif((cfg_clock_format   in 1 ..   5), cfg_clock_format  ,   5)

    SetVolume(cfg_volume)

    MACRO_ID = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':Id', _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    GotoBufferId(org_id)

    //  CLOCK_ID has two purposes:
    //  - To be as unique as possible in and across TSE and non-TSE sessions,
    //    so that it can be used as speak's uid parameter.
    //  - To be available as a temporary buffer for lists, etc.
    CLOCK_ID = (get_process_id() mod 100000) * 10000 + MACRO_ID mod 10000
    CLOCK_ID = CLOCK_ID

    Hook(_IDLE_        , idle)
    Hook(_NONEDIT_IDLE_, idle)
  endif
end WhenLoaded


proc Main()
  if voice_ok
    configure_main()
  endif
end Main


